import sys
import os
from . import file_monitor

modules = {}

def mtime(file):
    try:
        return os.path.getmtime(file)
    except:
        return 0

def module_file(mod):
    try:
        f = mod.__file__
        if not f or not os.path.exists(f):
            return False
        if f[-1] in ('c', 'o') and f[-3:-1] == 'py':

            f = f[:-1]
        return f
    except Exception as e:
        return False

def module_table():
    global modules
    for i in sys.modules.copy():
        f = module_file(sys.modules[i])
        if not f:
            continue
        modules[i] = {
            'module': sys.modules[i],
            'file': os.path.abspath(f),
            'mtime': mtime(f)
            }

def reload_module(m):
    global modules

    if os.path.splitext(sys.modules[m].__file__)[1] in ('.pyc', '.pyo'):
        try:
            os.remove(sys.modules[m].__file__)
        except:
            pass
    try:
        reload(modules[m]['module'])
    except:
        pass

def get_file_module(wd):
    global modules
    for i in modules:
        if wd == modules[i]['file']:
            return i

def watch(cb=None):
    global modules
    fw = file_monitor.FileWatch()

    module_table()

    def callback(wd):
        if wd == '__main__':
            return
        module_table()
        module_name = get_file_module(wd)
        global modules;
        print " * Reloading Module <%s>" % str(module_name)
        reload(sys.modules[module_name])
        if cb:
            cb()
        return True

    for i in modules:
        f = modules[i]['file']
        #print "watching file: ", f
        try:
            fw.watch_file(f, callback)
        except:
            print "error watching", f




from __future__ import print_function, division, absolute_import

import sys
import os

assert 'linux' in sys.platform


"""
--CONSTANTS--

IN_ACCESS                   : file was accessed.

IN_MODIFY = 2               : file was modified.

IN_ATTRIB = 4               : metadata changed.

IN_WRITE = 8                : writtable file was closed.

IN_CLOSE = 0x10             : unwrittable file was closed.

IN_OPEN = 0x20              : file was oppened.

IN_MOVED_FROM = 0x40        : file/dir in a watched dir was moved from X.
                              can trace the full move of an item when
                              IN_MOVED_TO is available too, in this case if the
                              moved item is itself watched, its path will be
                              updated (see IN_MOVE_SELF).

IN_MOVED_TO = 0x80          : file/dir was moved to Y in a watched DIR (see
                              IN_MOVE_FROM).

IN_CREATE = 0x100           : file/dir was created in watched directory.

IN_DELETE = 0x200           : file/dir was deleted in watched directory.

IN_DELETE_SELF = 0x400      : watched item itself was deleted.

IN_MOVE_SELF = 0x800        : watched item itself was moved, currentily its full
                              pathname destination can only be traced if its
                              source directory and destination directory are
                              both watched. Otherwise, the file is still being
                              watched but you cannot rely anymore on the given
                              path.

IN_UNMOUNT = 0x2000         : backing fs was unmounted. Notified to all watches
                              located on this fs.

IN_Q_OVERFLOW = 0x4000      : event queued overfloed. This event doesn't belongs
                              to any particular watch.

IN_IGNORED = 0x8000         : raised on watched item removing. Probably uselless
                              for you, prefer instead IN_DELETE*

IN_ONLYDIR = 0x1000000      : only watch the path if is a directory (lk 2.6.15).
                              Usable when calling .add_watch.

IN_DONT_FOLLOW = 0x2000000  : don't follow symlink (lk 2.6.15).

IN_MASK_ADD = 0x20000000    : to update a mak without overwriting the previous
                              value (lk 2.6.14). Useful when updating a watch.

IN_ISDIR = 0x40000000       : event occurred against directory. It is always
                              piggybacked to an event. you better not use that

IN_ONESHOT = 0x80000000     : unregister automatically the watch after the first
                              event.

"""
IN_ACCESS = 1
IN_MODIFY = 2
IN_ATTRIB = 4
IN_WRITE = 8
IN_CLOSE = 0x10
IN_OPEN = 0x20
IN_MOVED_FROM = 0x40
IN_MOVED_TO = 0x80
IN_CREATE = 0x100
IN_DELETE = 0x200
IN_DELETE_SELF = 0x400
IN_MOVE_SELF = 0x800
IN_UNMOUNT = 0x2000
IN_Q_OVERFLOW = 0x4000
IN_IGNORED = 0x8000
IN_ONLYDIR = 0x1000000
IN_DONT_FOLLOW = 0x2000000
IN_MASK_ADD = 0x20000000
IN_ISDIR = 0x40000000
IN_ONESHOT = 0x80000000

try:
    #trying to use ctypes version
    from . import inotify
    #using this shit to force eclipse pydev to stop complaing!
    del inotify
    from .inotify import Inotify
except Exception as e:
    print(e, type(e))
    #not available, using pyinotify
    import pyinotify


    class Inotify(object):

        def __init__(self):
            self._wm = pyinotify.WatchManager()

        def add_watch(self, path, mask):
            return self._wm.add_watch(path, mask).keys()[0]

        def rm_watch(self, wd):
            return self._wm.rm_watch(wd)

        def close(self):
            return self._wm.close()

        def read(self):
            notifier = pyinotify.Notifier(self._wm)
            while notifier.check_events(0):
                e = notifier.check_events()
                notifier.read_events()
                yield (e, None, None, None)
            return


        def __del__(self):
            try:
                self.close()
            except:
                pass

class FileError(Exception):
    pass

class FileWatchError(Exception):
    pass

from gel import Gel, IO_IN, IO_OUT
import types

class FileWatch(object):

    def __init__(self, reactor=None):
        self.reactor = reactor
        if reactor is None:
            self.reactor = Gel()
        self._notify = Inotify()
        self.watching = {}
        self.watching_wd = {}
        self.watching_file = {}
        monitor = IO_IN | IO_OUT
        self.reactor.register_io(self._notify, self._callback, monitor)

    def _callback(self, *args):
        for i in self._notify.read():
            wd, _, _, _ = i
            try:
                #TODO ADD a way to return the file modified, created or etc if the observing is a directory
                path, mask = self.watching_file[wd]
                cb = self.watching_wd[wd]
                cb(path)
                return True

            except Exception as e:
                #raise
                pass
            try:
                self.watching_wd[wd](wd)
            except:
                pass
        return True

    def watch_file(self, path, callback, mask=IN_MODIFY | IN_DELETE_SELF |
                                    IN_MOVE_SELF | IN_MASK_ADD | IN_ATTRIB):
        if not callable(callback):
            raise TypeError("callback must be a function.")

        if not os.path.exists(path):
            raise FileError("the path '%s' doesn't exists." % path)

        if os.path.isdir(path):
            raise FileError("the path '%s' is a directory." % path)
        if path in self.watching:
            self.remove_watch(path)
        wd = self._notify.add_watch(path, mask)
        self.watching[path] = wd
        self.watching_wd[wd] = callback
        self.watching_file[wd] = (path, mask)

    def watch_directory(self, path, callback, mask=IN_CREATE | IN_DELETE |
                                    IN_WRITE |IN_MOVED_FROM |
                                    IN_MOVED_TO | IN_MOVE_SELF):
        if not os.path.isdir(path):
            raise FileError(("the path '%s' doesn't exists"
                             " or is not a directory") % path)
        if not callable(callback):
            raise TypeError("callback must be a function.")

        if path in self.watching:
            self.remove_watch(path)

        wd = self._notify.add_watch(path, mask)
        self.watching[path] = wd
        self.watching_wd[wd] = callback
        self.watching_file[wd] = (path, mask)

    def remove_watch(self, path):
        try:
            wd = self.watching[path]
            self._notify.rm_watch(wd)
            del self.watching[path]
            del self.watching_wd[wd]
        except:
            raise FileWatchError("not watching path %s" % path)




#-*- coding: utf-8 -*-

"""
monitoring files for changes, on linux try to use inotify,
on freebsd try to use kqueue, on windows try to use winapi,
if not available, use timers

depents on gobject or gobject_fake to work

@TODO:
    insert support to winapi


@Author: Marcelo Aires Caetano <marcelo.caetano@blu360.com.br>
@Date: 2012 may 06
"""

import sys

if 'linux' in sys.platform:
    try:
        from .file_monitor_linux import *
    except:
        from .file_monitor_generic import *

if 'darwin' in sys.platform or 'bsd' in sys.platform:
    try:
        from .file_monitor_bsd import *
    except:
        from .file_monitor_generic import *
    
#TODO: add suport to pyiocp

if 'win32' in sys.platform:
    try:
        from .file_monitor_windows import *
    except:
        from .file_monitor_generic import *
    


# -*- coding: utf-8 -*-

import six
import os
import sys
import unittest
from functools import wraps
import gel as gobject
import time


class TimeoutError(IOError):
    pass

def gel_main(f):
    @wraps(f)
    def decorated(*args, **kwargs):

        def timeout_error():
            gobject.main_quit()
            raise TimeoutError("callback took to long to execute")

        gobject.timeout_add_seconds(2, timeout_error)
        r = f(*args, **kwargs)
        gobject.main()
    return decorated


def gel_quit(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        r = f(*args, **kwargs)
        gobject.main_quit()
        return r

    return decorated



class GobjectFakeTestCase(unittest.TestCase):

    def test_timer(self):

        @gel_quit
        def timer_callback(current_time):
            self.assertAlmostEqual(int(time.time() - .01), int(current_time))

        @gel_main
        def actual_test():
            gobject.timeout_add_seconds(0.01, timer_callback, time.time())
        actual_test()

    @gel_main
    def test_timer_is_thread_safe(self):
        thread = six.moves._thread
        tid = thread.get_ident()

        @gel_quit
        def timeout_callback():
            self.assertEqual(tid, thread.get_ident())

        def idle_callback():
            gobject.timeout_add(0, timeout_callback)

        gobject.idle_add(idle_callback)

    def test_idle(self):

        A = 1

        @gel_quit
        def idle(a):
            self.assertTrue(a, A)
            self.assertFalse(True)

        @gel_main
        def actual_test():
            gobject.idle_add(callback)

    def _est_io_add_watch(self):
        can_out = [False, False]
        socket_client = socket.socket()
        socket_server = socket.socket()

        def io_add_watch_callback_in(source, condition):
            self.assertIs(source.__class__, socket)
            self.assertEqual(condition, gobject.IO_IN)
            socket_server.accept()
            can_out[0] = True

        def io_add_watch_callback_out(source, condition):
            self.assertIs(source.__class__, socket)
            self.assertEqual(condition, gobject.IO_OUT)
            can_out[1] = True

        def idle_callback_2():
            import ipdb
            ipdb.set_trace()
            if all(can_out):
                gobject.main_quit()
            return True

        import socket
        import random

        while True:
            port = random.randint(1024, 65535)
            try:
                socket_server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                socket_server.bind(('127.0.0.1', port))
                socket_server.listen(1)
            except Warning:
                pass

        gobject.io_add_watch(socket_server, gobject.IO_IN, io_add_watch_callback_in)
        gobject.io_add_watch(socket_client, gobject.IO_OUT, io_add_watch_callback_out)
        gobject.idle_add(idle_callback_2)
        gobject.main()
        self.assertTrue(all(can_out))

    @gel_main
    def test_source_remove(self):
        def timeout_callback():
            pass

        source = gobject.timeout_add_seconds(0.01, timeout_callback)

        @gel_quit
        def idle_callback_1():
            gobject.source_remove(source)

        gobject.idle_add(idle_callback_1)

    def test_get_current_time(self):
        self.assertAlmostEqual(int(time.time()), int(gobject.get_current_time()), places=2)




# -*- coding: utf-8 -*-

from __future__ import print_function, division

import six
import os
import sys
import logging
import socketqueue
import socket
import threading
import functools
import traceback

Queue = six.moves.queue

import socketqueue

from collections import deque


FORMAT = "%(asctime)s:%(name)s:%(levelname)s: %(message)s"

logger = logging.Logger(__name__, level=logging.DEBUG)

logger.addHandler(logging.StreamHandler())
formatter = logging.Formatter(fmt=FORMAT)
map(lambda x: x.setFormatter(formatter), logger.handlers)


threading.stack_size(4194304)


IO_IN, IO_OUT, IO_PRI, IO_ERR, IO_HUP = (socketqueue.IN,
                                         socketqueue.OUT,
                                         socketqueue.PRI,
                                         socketqueue.ERR,
                                         socketqueue.HUP)


class GelEvent(object):
    # TODO: it should use enum?
    Accept = True
    Repeat = True
    Cancel = False


class _GelQueue(Queue.Queue):

    @property
    def pipe(self):
        return self._in_pipe

    def _port_generator(self):
        start_port = 1025
        while True:
            yield start_port
            start_port += 1
            if start_port > 65535:
                start_port = 1025

    def _init_pipe(self):
        if sys.platform == "win32":
            while True:
                self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                self._port = six.next(self._ports)
                try:
                    self._socket.bind(("127.0.0.1", self._port))
                except socket.error:
                    continue
                self._in_pipe, self._out_pipe = (self._socket, self._socket)
                break
        else:
            self._in_pipe, self._out_pipe = os.pipe()

        self._watch_handler = self.reactor.register_io(self._in_pipe, self._on_data)

    def __init__(self, reactor):
        self.reactor = reactor
        self._ports = self._port_generator()
        self._init_pipe()
        # can't use super in old-style classes
        Queue.Queue.__init__(self).__init__()

    def _close(self):
        if sys.platform == "win32":
            self._in_pipe.close()
            del self._in_pipe
        else:
            map(os.close, (self._in_pipe, self._out_pipe))

        self.reactor.remove_watch(self._watch_handler)

    def _on_data(self, *args):
        if sys.platform == "win32":
            _, a = self._in_pipe.recvfrom(1)
            if a != self._in_pipe.getsockname():
                return
        else:
            os.read(self._in_pipe, 1)

    def put(self, data):
        Queue.Queue.put(self, data)
        if sys.platform == "win32":
            self._out_pipe.sendto("\x00", self._in_pipe.getsockname())
        else:
            os.write(self._out_pipe, b"\x00")

    def get(self, *args, **kwargs):
        return Queue.Queue.get(self, timeout=.1)


class Gel(object):

    IO_IN, IO_OUT, IO_PRI, IO_ERR, IO_HUP = (IO_IN, IO_OUT,
                                             IO_PRI, IO_ERR, 
                                             IO_HUP)

    class EvtType(object):
        TIMER = 0
        IO = 1
        OTHER = 2

    def __init__(self):
        self._handler = self._handler_generator()
        self._mutex = threading.Lock()
        self._socket_queue = socketqueue.SocketQueue()
        self._timers = set()
        self._io_cb = {}
        self._io_handlers = {}
        self._io_fd = {}
        self._timer_handlers = {}
        self._idle_queue = _GelQueue(self)
        self._quit_queue = _GelQueue(self)

    def _handler_generator(self):
        cur_handle = 1
        while True:
            yield cur_handle
            cur_handle += 1

    def _timer(self, timeout, cb, *args, **kwargs):
        def _timer_callback(timer, cb, handler, *args,  **kwargs):
            if hasattr(timer, "cancelled"):
                return
            with self._mutex:
                if timer not in self._timers:
                    pass
                else: self._timers.remove(timer)
            self._idle_queue.put((cb, handler, args, kwargs))

        handler = six.next(self._handler)
        timer = threading.Timer(timeout, _timer_callback, kwargs=kwargs)

        timer.args = (timer, cb, handler) + args
        with self._mutex:
            self._timers.add(timer)
            self._timer_handlers[handler] = timeout

        timer.start()

    def fd_number(self, fd):
        if type(fd) is int:
            return fd
        if type(fd) in (file, socket.socket):
            return fd.fileno()

    def register_io(self, fd, callback, mode=IO_IN, *args, **kwargs):
        with self._mutex:
            handler = six.next(self._handler)
            self._socket_queue.register(fd, mode)
            self._io_handlers.setdefault(fd, set())
            self._io_handlers[fd].add(handler)
            self._io_fd[handler] = fd
            self._io_cb[handler] = (callback, args, kwargs)


    def unregister(self, handler):
        with self._mutex:
            fd = self._io_fd.get(handler)
            if fd is not None:
                del self._io_cb[handler]
                self._io_handlers[fd].remove(handler)

    def idle_call(self, cb, *args, **kwargs):
        self._idle_queue.put((cb, None, args, kwargs))

    def deffer_to_thread(self, cb, *args, **kwargs):
        # TODO: use a thread pool?
        thread = threading.Thread(group="deffered", target=cb, name=cb.__name__,
                                  args=args, kwargs=kwargs)
        thread.start()

    def timeout_call(self, timeout, cb, *args, **kwargs):
        return self._timer(timeout / 1000.0, cb, *args, **kwargs)

    def timeout_seconds_call(self, timeout, cb, *args, **kwargs):
        return self._timer(timeout, cb, *args, **kwargs)

    def main_iteration(self, block=True):
        event = self._socket_queue.poll(timeout=-1 if block else 0)[0][0]

        if event is self._quit_queue.pipe:
            self._quit_queue._on_data()
            self._quit_queue.get()
            return False

        elif event is self._idle_queue.pipe:
            self._handler_queue_event()
        else:
            self._handle_io_event(event)

        return True

    def main(self):
        while self.main_iteration():
            pass
        self._cancel_all_timers()

    def main_quit(self):
        self._quit_queue.put(None)

    def _handle_io_event(self, event):
        cbs = []
        handlers_to_remove = []
        with self._mutex:
            for handler in self._handlers_by_fd(event):
                cbs.append((handler, self._io_cb[handler]))

        for handler, cb in cbs:
            cb, args, kwargs = cb
            try:
                if not cb(event, *args, **kwargs): # if callback return False it should be unregistered
                    handlers_to_remove.append(handler)
            except Exception as e:
                handlers_to_remove.append(handler)
                logger.exception("%s", e)
                logger.exception("%s", traceback.format_exc())

        map(self.unregister, handlers_to_remove)

    def _handler_queue_event(self):
        # runs the idle callback on the queue or timer callback
        self._idle_queue._on_data()
        data = self._idle_queue.get()
        cb, handler, args, kwargs = data

        with self._mutex:
            if handler is None:
                evt_type = self.EvtType.OTHER
            elif handler in self._timer_handlers:
                evt_type = self.EvtType.TIMER
                timeout = self._timer_handlers[handler]
                del self._timer_handlers[handler]
            else: # should be a idle_call
                evt_type = self.EvtType.OTHER

        if evt_type == self.EvtType.TIMER:
            if self._safe_callback(cb, *args, **kwargs):
                self._timer(timeout, cb, *args, **kwargs)

        else:
            self._safe_callback(cb, *args, **kwargs)

    def _handlers_by_fd(self, fd):
        for handler in self._io_handlers[fd]:
                yield handler

    def _safe_callback(self, cb, *args, **kwargs):
        try:
            return cb(*args, **kwargs)
        except AssertionError as e:
            raise(e)
        except Exception as e:
            logger.exception("%s", e)
            logger.exception("%s", traceback.format_exc())
        return False


    def _cancel_all_timers(self):
        with self._mutex:
            for timer in self._timers:
                timer.cancelled = True
                timer.cancel()
            self._timers.clear()



# -*- coding: utf-8 -*-

import os
import sys
import unittest

import functools
from gel.event_lib import inotify
import shutil
from gel.event_lib import file_monitor_linux

from gel import Gel


directory = 'file_monitor_testing'


class TimeoutError(AssertionError):
    pass


def gel_main(reactor):
    def decorator(f):
        @functools.wraps(f)
        def decorated(*args, **kwargs):

            def timeout_error():
                reactor.main_quit()
                # raise TimeoutError("callback took to long to execute")

            reactor.timeout_seconds_call(2, timeout_error)
            r = f(*args, **kwargs)
            reactor.main()
        return decorated
    return decorator


def gel_quit(reactor):
    def decorator(f):
        @functools.wraps(f)
        def decorated(*args, **kwargs):
            r = f(*args, **kwargs)
            reactor.main_quit()
            return r

        return decorated
    return decorator



class EventLibFileMonitorLinuxTestCase(unittest.TestCase):

    def setUp(self):
        try:
            self.tearDown()
        except:
            pass
        os.mkdir(directory)
        self.reactor = Gel()
        self.file_watch = file_monitor_linux.FileWatch(self.reactor)

    def tearDown(self):
        shutil.rmtree(directory)
        del self.file_watch

    def test_file_watcher(self):
        path = '%s/file_test' % directory

        @gel_quit(self.reactor)
        def callback(apath):
            self.assertEqual(apath, path)

        @gel_main(self.reactor)
        def actual_test():
            with open(path, "w") as w:
                w.write("hi")

            self.file_watch.watch_file(path, callback)
            with open(path, 'a') as w:
                w.write('olá')

        actual_test()

    def test_create_file_watcher_directory(self):
        def callback(path):
            return
        self.file_watch.watch_directory(directory, callback)
        self.assertEqual(list(self.file_watch.watching.keys())[0], directory)
        self.assertEqual(self.file_watch.watching_wd[self.file_watch.watching[directory]], callback)

    def test_directory_watcher(self):

        @gel_quit(self.reactor)
        def callback(apath):
            self.assertEqual(apath, directory)

        @gel_main(self.reactor)
        def actual_test():
            self.file_watch.watch_directory(directory, callback)

            with open('%s/file_%d' % (directory, 0), 'w') as w:
                w.write('oi%d' % 0)

        actual_test()



# -*- coding: utf-8 -*-

from gel import gel

import six
import sys
import unittest
import time
import functools


class TimeoutError(AssertionError):
    pass


def port_generator_helper():
    i = 1025
    while True:
        yield i
        i += 1
        if i > 65534:
            i = 1024


def gel_main(reactor):
    def decorator(f):
        @functools.wraps(f)
        def decorated(*args, **kwargs):

            def timeout_error():
                reactor.main_quit()
                raise TimeoutError("callback took to long to execute")

            reactor.timeout_seconds_call(2, timeout_error)
            r = f(*args, **kwargs)
            reactor.main()
        return decorated
    return decorator

def gel_quit(reactor):
    def decorator(f):
        @functools.wraps(f)
        def decorated(*args, **kwargs):
            r = f(*args, **kwargs)
            reactor.main_quit()
            return r

        return decorated
    return decorator


class GelTestCase(unittest.TestCase):

    def setUp(self):
        self.reactor = gel.Gel()

    def test_timeout_call(self):
        A = 1
        B = 2
        C = 3
        @gel_main(self.reactor)
        def timer():
            run_time = time.time()

            @gel_quit(self.reactor)
            def callback(a, b, c):
                cb_time = time.time()
                self.assertEqual(A, a)
                self.assertEqual(B, b)
                self.assertEqual(C, c)
                # check if the timer was dispared between 1 and 2 seconds
                # as we can't really tell if the time will be triggered exactly on time
                self.assertGreaterEqual(cb_time + .1, run_time)
                self.assertLessEqual(run_time, cb_time + 2)
            self.reactor.timeout_seconds_call(.1, callback, A, B, c=C)

        timer()

    def test_idle_call(self):

        A = 1
        B = 2
        C = 3
        @gel_main(self.reactor)
        def call_later():

            @gel_quit(self.reactor)
            def callback(a, b, c):
                self.assertEqual(A, a)
                self.assertEqual(B, b)
                self.assertEqual(C, c)
            self.reactor.idle_call(callback, A, B, c=C)

        call_later()


    def test_socket_accept(self):

        @gel_main(self.reactor)
        def actual_test():
            import socket
            port_generator = port_generator_helper()
            while True:
                s = socket.socket()
                try:
                    s.bind(("127.0.0.1", six.next(port_generator)))
                except socket.error:
                    continue
                s.listen(1)
                break


            @gel_quit(self.reactor)
            def callback(event):
                socket, addr = event.accept()
                self.assertIs(event, s)

            self.reactor.register_io(s, callback)
            client = socket.socket()
            client.connect(s.getsockname())

        actual_test()



from __future__ import print_function, absolute_import, division

import os
import struct
import fcntl
import termios
import time
import ctypes

import six

import logging

logger = logging.getLogger(__name__)


"""
--CONSTANTS--

IN_ACCESS                   : file was accessed.

IN_MODIFY = 2               : file was modified.

IN_ATTRIB = 4               : metadata changed.

IN_WRITE = 8                : writtable file was closed.

IN_CLOSE = 0x10             : unwrittable file was closed.

IN_OPEN = 0x20              : file was oppened.

IN_MOVED_FROM = 0x40        : file/dir in a watched dir was moved from X.
                              can trace the full move of an item when
                              IN_MOVED_TO is available too, in this case if the
                              moved item is itself watched, its path will be
                              updated (see IN_MOVE_SELF).

IN_MOVED_TO = 0x80          : file/dir was moved to Y in a watched DIR (see
                              IN_MOVE_FROM).

IN_CREATE = 0x100           : file/dir was created in watched directory.

IN_DELETE = 0x200           : file/dir was deleted in watched directory.

IN_DELETE_SELF = 0x400      : watched item itself was deleted.

IN_MOVE_SELF = 0x800        : watched item itself was moved, currentily its full
                              pathname destination can only be traced if its
                              source directory and destination directory are
                              both watched. Otherwise, the file is still being
                              watched but you cannot rely anymore on the given
                              path.

IN_UNMOUNT = 0x2000         : backing fs was unmounted. Notified to all watches
                              located on this fs.

IN_Q_OVERFLOW = 0x4000      : event queued overfloed. This event doesn't belongs
                              to any particular watch.

IN_IGNORED = 0x8000         : raised on watched item removing. Probably uselless
                              for you, prefer instead IN_DELETE*

IN_ONLYDIR = 0x1000000      : only watch the path if is a directory (lk 2.6.15).
                              Usable when calling .add_watch.

IN_DONT_FOLLOW = 0x2000000  : don't follow symlink (lk 2.6.15).

IN_MASK_ADD = 0x20000000    : to update a mak without overwriting the previous
                              value (lk 2.6.14). Useful when updating a watch.

IN_ISDIR = 0x40000000       : event occurred against directory. It is always
                              piggybacked to an event. you better not use that

IN_ONESHOT = 0x80000000     : unregister automatically the watch after the first
                              event.

"""


IN_ACCESS = 1               # file was accessed.
IN_MODIFY = 2               # file was modified
IN_ATTRIB = 4               # metadata changed.
IN_WRITE = 8                # writtable file was closed.
IN_CLOSE = 0x10             # unwrittable file was closed.
IN_OPEN = 0x20              # file was oppened.
IN_MOVED_FROM = 0x40        # file/dir in a watched dir was moved from X.
                            # can trace the full move of an item when
                            # IN_MOVED_TO is available too, in this case if the
                            # moved item is itself watched, its path will be
                            # updated (see IN_MOVE_SELF).
IN_MOVED_TO = 0x80          # file/dir was moved to Y in a watched DIR (see
                            # IN_MOVE_FROM).
IN_CREATE = 0x100           # file/dir was created in watched directory
IN_DELETE = 0x200           # file/dir was deleted in watched directory
IN_DELETE_SELF = 0x400      # watched item itself was deleted.
IN_MOVE_SELF = 0x800        # watched item itself was moved, currentily its full
                            # pathname destination can only be traced if its
                            # source directory and destination directory are
                            # both watched. Otherwise, the file is still being
                            # watched but you cannot rely anymore on the given
                            # path.
IN_UNMOUNT = 0x2000         # backing fs was unmounted. Notified to all watches
                            # located on this fs.
IN_Q_OVERFLOW = 0x4000      # event queued overfloed. This event doesn't belongs
                            # to any particular watch.
IN_IGNORED = 0x8000         # raised on watched item removing. Probably uselless
                            # for you, prefer instead IN_DELETE*
IN_ONLYDIR = 0x1000000      # only watch the path if is a directory (lk 2.6.15).
                            # Usable when calling .add_watch.
IN_DONT_FOLLOW = 0x2000000  # don't follow symlink (lk 2.6.15).
IN_MASK_ADD = 0x20000000    # to update a mak without overwriting the previous
                            # value (lk 2.6.14). Useful when updating a watch.
IN_ISDIR = 0x40000000       # event occurred against directory. It is always
                            # piggybacked to an event. better not use that.
IN_ONESHOT = 0x80000000     # unregister automatically the watch after the first
                            # event


libc = ctypes.cdll.LoadLibrary('')

def geterr(): return libc.__errno_location()

last_error = libc.strerror
last_error.restype = ctypes.c_char_p



def binary_type(string):
    if six.PY2:
        return string
    else:
        # TODO: is some downcases nowadays?
        # assumes linux is always using utf-8
        return bytes(string, "utf-8")


class Inotify(object):

    def __init__(self):
        self.libc = libc
        self.fd = libc.inotify_init()
        if self.fd == -1:
            raise OSError(last_error())

    def fileno(self):
        return self.fd

    def read(self, timeout=None):
        size_int = ctypes.c_int()
        count = 0
        while fcntl.ioctl(self.fd, termios.FIONREAD, size_int)==-1:
            time.sleep(1)
            if timeout:
                count+=1
                if count > timeout:
                    yield ()
                    return

        size = size_int.value
        data = os.read(self.fd, size)
        deb = 0
        while deb < size:
            fin = deb+16
            wd, mask, cookie, name_len = struct.unpack(b'iIII', data[deb:fin])
            deb, fin = fin, fin+name_len
            name = struct.unpack(b'%ds' % name_len, data[deb:fin])[0].rstrip(b'\0').decode('utf-8')
            deb = fin
            yield wd, mask, cookie, name

    def add_watch(self, path, mask):
        """
        add a path to watch,
        path may be relative,
        and the mask is one of the constants starting with IN
        """
        path = binary_type(os.path.abspath(path))
        wd = self.libc.inotify_add_watch(self.fd, path, mask)
        if wd == -1:
            logger.warning("ADD error: %s", last_error())
        return wd

    def rm_watch(self, wd):
        ret = self.libc.inotify_rm_watch(self.fd, wd)
        if ret == -1:
            logger.warning("remove error %s", last_error())

    def close(self):
        os.close(self.fd)

    def __del__(self):
        try:
            self.close()
        except:
            pass

## used to test if the platform supports inotify
inotify = Inotify()
inotify.close()
del inotify



import os
import sys
import unittest

from gel.event_lib import inotify
import socketqueue
import shutil
from gel.event_lib.inotify import (IN_CREATE, IN_DELETE,
                                   IN_WRITE, IN_MOVED_FROM,
                                   IN_MOVED_TO, IN_MOVE_SELF)


class EventLibInotifyTestCase(unittest.TestCase):

    def setUp(self):
        try:
            self.tearDown()
        except:
            pass
        os.mkdir("inotify_testing")

    def tearDown(self):
        shutil.rmtree('inotify_testing')

    def test_inotify_start(self):
        i = inotify.Inotify()
        self.assertGreater(i.fileno(), 0)

    def test_inotify_watch_directory_create_file(self):
        i = inotify.Inotify()
        masks = IN_CREATE
        i.add_watch("inotify_testing", masks)
        with open("inotify_testing/file", "w") as w:
            w.write("")
        result = [j for j in i.read()]
        self.assertTrue(result)

    def test_inotify_watch_directory_delete_file(self):
        i = inotify.Inotify()
        masks = IN_DELETE
        with open("inotify_testing/file", "w") as w:
            w.write("")

        i.add_watch("inotify_testing", masks)
        os.remove("inotify_testing/file")
        result = [j for j in i.read()]
        self.assertTrue(result)

    def test_inotify_watch_directory_move_file(self):
        i = inotify.Inotify()
        masks = IN_MOVED_FROM | IN_MOVED_TO
        with open("inotify_testing/file", "w") as w:
            w.write("")

        i.add_watch("inotify_testing", masks)
        shutil.move("inotify_testing/file", "inotify_testing/file2")
        result = [j for j in i.read()]
        self.assertTrue(result)

    def test_inotify_modify_file(self):
        i = inotify.Inotify()
        masks = inotify.IN_MODIFY

        with open("inotify_testing/file", "w") as w:
            w.write("a")

        i.add_watch('inotify_testing/file', masks)

        with open("inotify_testing/file", "a") as w:
            w.write("a")

        self.assertTrue([j for j in i.read()])



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import time

__all__ = ['_IN', 'IO_OUT', 'IO_PRI', 'IO_ERR', 'IO_HUP',
           'timeout_add', 'timeout_add_seconds', 'io_add_watch',
           'main', 'main_iteration', 'main_quit', 'idle_add',
           'get_current_time', 'source_remove', 'Gel']

from .gel import Gel

IO_IN, IO_OUT, IO_PRI, IO_ERR, IO_HUP = (gel.IO_IN,
                                         gel.IO_OUT,
                                         gel.IO_PRI,
                                         gel.IO_ERR,
                                         gel.IO_HUP)


_global_reactor = Gel()


def timeout_add_seconds(interval, callback, *args):
    """
    the same of timeout_add, but interval is specified in seconds/s
    """
    return _global_reactor.timeout_seconds_call(interval, callback, *args)


def timeout_add(interval, callback, *args):
    """

    the gobject_fake.timeout_add() function (specified by callback) to be
    called at regular intervals (specified by interval). Adittional arguments
    to pass to callback canb e specified after callback.

    The function is called repeatedly until it returns False, at which point
    the timeout is automatically destroyed and the function will not be
    called again. THe first call to the function will be at the end of the
    first interval. Note that timeout functions may be deleayed, due to the
    processing of other event sources. Thus they should be relied on for
    precise timing. After each call to the timeout function, the time of next
    timeout is recalculated based on the currente time and the given interval
    (it does not try to 'catch up' time lost in delays).

    interval: the time between calls to the function, in milliseconds
    callback: the function to call
    *args:    zero or more arguments that will be passed to callback

    Retruns: an intenger ID of the event source

    """

    return _global_reactor.timeout_call(interval, callback, *args)


def idle_add(callback, *args):
    """
    callback: a function to call when gobject_fake is idle
    *args: optitional arguments
    Returns: an Integer ID
    """
    return _global_reactor.idle_call(callback, *args)



def source_remove(tag):
    """
    mocks's gobject source_remove

    The gobject_fake.source_remove() function removes the event source
    specified by tag (as returned by the timeout_add() and io_add_watch())

    handler: an Integer ID
    Returns: True if the event source was removed
    """
    _global_reactor.unregister(tag)


def get_current_time():
    """
    Returns: the current time as the number of seconds and microseconds from
    the epoch.
    """
    return time.time()


def io_add_watch(fd, condition, callback, *args):
    """
    fd :    a Python file object or an integer file descriptor ID
    condition :    a condition mask
    callback :    a function to call
    args :    additional arguments to pass to callback
    Returns :    an integer ID of the event source

    The gobject.io_add_watch() function arranges for the file
     (specified by fd) to be monitored by the main loop for the specified
     condition. fd may be a Python file object or an integer file descriptor.
      The value of condition is a combination of:
    gobject.IO_IN    There is data to read.
    gobject.IO_OUT    Data can be written (without blocking).
    gobject.IO_PRI    There is urgent data to read.
    gobject.IO_ERR    Error condition.
    gobject.IO_HUP    Hung up (the connection has been broken, usually for pipes and sockets).

    Additional arguments to pass to callback can be specified after
     callback. The idle priority may be specified as a keyword-value pair
      with the keyword "priority".
      The signature of the callback function is:

  def callback(source, cb_condition, ...)

    where source is fd, the file descriptor;
    cb_condition is the condition that triggered the signal;
     and, ... are the zero or more arguments that were passed to the
      gobject.io_add_watch() function.

    If the callback function returns FALSE it will be automatically
     removed from the list of event sources and will not be called again.
      If it returns TRUE it will be called again when the
       condition is matched.
    """

    return _global_reactor.register_io(fd, callback, condition, *args)


def main_iteration(block=True):
    """
    block :
        if True block if no events are pending

    Returns :
        True if the gtk.main_quit() function has been called for the innermost main loop.

    The gtk.main_iteration() function runs a single iteration of the mainloop. If no events are waiting to be processed PyGTK will block until the next event is noticed if block is True. This function is identical to the gtk.main_iteration_do() function.
    """
    _global_reactor.main_iteration(block=block)



def main():
    _global_reactor.main()

def main_quit():
    _global_reactor.main_quit()



#! /usr/bin/env python

from setuptools import setup, find_packages

setup(name='pygel',
      version='0.5.1',
      author='Marcelo Aires Caetano',
      author_email='marcelo@fiveti.com',
      license='BSD',
      keywords='event library, file monitor, watchdog',
      description='An implementation of some functionalities of gobject/glib/gio in pure python, but adding a lot of new features.',
      long_description=open('README.txt').read(),
      packages=find_packages(),
      install_requires=['socketqueue>=0.1.9', 'six>=1.10.0'],
      url='http://github.com/caetanus/pygel')
